@@ -26,6 +26,9 @@
 #ifndef CGAL_MESH_3_MESH_COMPLEX_3_IN_TRIANGULATION_3_BASE_H
 #define CGAL_MESH_3_MESH_COMPLEX_3_IN_TRIANGULATION_3_BASE_H
 
+#include <CGAL/license/Mesh_3.h>
+
+
 #include <CGAL/Mesh_3/config.h>
 
 #include <CGAL/Mesh_3/utilities.h>
@@ -38,8 +41,49 @@
 #include <CGAL/Mesh_3/io_signature.h>
 #include <CGAL/Union_find.h>
 
+#include <boost/functional/hash.hpp>
+
 #ifdef CGAL_LINKED_WITH_TBB
   #include <tbb/atomic.h>
+  #include <tbb/concurrent_hash_map.h>
+
+namespace CGAL {
+  template < class DSC, bool Const >
+  std::size_t tbb_hasher(const CGAL::internal::CC_iterator<DSC, Const>& it)
+  {
+    return CGAL::internal::hash_value(it);
+  }
+
+
+  template < class DSC, bool Const >
+  std::size_t tbb_hasher(const CGAL::CCC_internal::CCC_iterator<DSC, Const>& it)
+  {
+    return CGAL::CCC_internal::hash_value(it);
+  }
+
+  // As Marc Glisse pointed out the TBB hash of a std::pair is
+  // simplistic and leads to the
+  // TBB Warning: Performance is not optimal because the hash function
+  //              produces bad randomness in lower bits in class
+  //              tbb::interface5::concurrent_hash_map
+  template < class DSC, bool Const >
+  std::size_t tbb_hasher(const std::pair<CGAL::internal::CC_iterator<DSC, Const>,
+                                         CGAL::internal::CC_iterator<DSC, Const> >& p)
+  {
+    return boost::hash<std::pair<CGAL::internal::CC_iterator<DSC, Const>,
+                                 CGAL::internal::CC_iterator<DSC, Const> > >()(p);
+  }
+
+
+  template < class DSC, bool Const >
+  std::size_t tbb_hasher(const std::pair<CGAL::CCC_internal::CCC_iterator<DSC, Const>,
+                                         CGAL::CCC_internal::CCC_iterator<DSC, Const> >& p)
+  {
+    return boost::hash<std::pair<CGAL::CCC_internal::CCC_iterator<DSC, Const>,
+                                 CGAL::CCC_internal::CCC_iterator<DSC, Const> > >()(p);
+  }
+
+}
 #endif
 
 namespace CGAL {
@@ -167,9 +211,8 @@ public:
   void clear() {
     number_of_cells_ = 0;
     number_of_facets_ = 0;
+    clear_manifold_info();
     tr_.clear();
-    manifold_info_initialized_ = false;
-    edge_facet_counter_.clear();
   }
 
   /// Assignment operator
@@ -231,7 +274,11 @@ public:
     //test incident edges for REGULARITY and count BOUNDARY edges
     typename std::vector<Edge> edges;
     edges.reserve(64);
-    tr_.incident_edges(v, std::back_inserter(edges));
+    if(tr_.is_parallel()) {
+      tr_.incident_edges_threadsafe(v, std::back_inserter(edges));
+    } else {
+      tr_.incident_edges(v, std::back_inserter(edges));
+    }
     int number_of_boundary_incident_edges = 0; // could be a bool
     for (typename std::vector<Edge>::iterator
            eit=edges.begin(), end = edges.end();
@@ -282,7 +329,15 @@ public:
   {
     if(!manifold_info_initialized_) init_manifold_info();
 
+#ifdef CGAL_LINKED_WITH_TBB
+    typename Edge_facet_counter::const_accessor accessor;
+    if(!edge_facet_counter_.find(accessor,
+				 this->make_ordered_pair(edge)))
+      return NOT_IN_COMPLEX;
+    switch(accessor->second)
+#else // not CGAL_LINKED_WITH_TBB
     switch(edge_facet_counter_[this->make_ordered_pair(edge)])
+#endif // not CGAL_LINKED_WITH_TBB
     {
     case 0: return NOT_IN_COMPLEX;
     case 1: return BOUNDARY;
@@ -467,6 +522,29 @@ public:
   /// Returns bbox
   Bbox_3 bbox() const;
 
+  void clear_cells_and_facets_from_c3t3() {
+    for(typename Tr::Finite_cells_iterator
+          cit = this->triangulation().finite_cells_begin(),
+          end = this->triangulation().finite_cells_end();
+        cit != end; ++cit)
+    {
+      set_subdomain_index(cit, Subdomain_index());
+    }
+    this->number_of_cells_ = 0;
+    for(typename Tr::Finite_facets_iterator
+          fit = this->triangulation().finite_facets_begin(),
+          end = this->triangulation().finite_facets_end();
+        fit != end; ++fit)
+    {
+      Facet facet = *fit;
+      Facet mirror = tr_.mirror_facet(facet);
+      set_surface_patch_index(facet.first, facet.second, Surface_patch_index());
+      set_surface_patch_index(mirror.first, mirror.second, Surface_patch_index());
+    }
+    this->number_of_facets_ = 0;
+    clear_manifold_info();
+  }
+
   void clear_manifold_info() {
     edge_facet_counter_.clear();
     manifold_info_initialized_ = false;
@@ -498,7 +576,16 @@ private:
           const int edge_index_vb = tr_.vertex_triple_index(i, (j == 2) ? 0 : (j+1));
           const Vertex_handle edge_va = cell->vertex(edge_index_va);
           const Vertex_handle edge_vb = cell->vertex(edge_index_vb);
+#ifndef CGAL_LINKED_WITH_TBB
           ++edge_facet_counter_[this->make_ordered_pair(edge_va, edge_vb)];
+#else // CGAL_LINKED_WITH_TBB
+	  {
+	    typename Edge_facet_counter::accessor accessor;
+	    edge_facet_counter_.insert(accessor,
+				       this->make_ordered_pair(edge_va, edge_vb));
+	    ++accessor->second;
+	  }
+#endif // CGAL_LINKED_WITH_TBB
 
           const std::size_t n = edge_va->cached_number_of_incident_facets();
           edge_va->set_c2t3_cache(n+1, -1);
@@ -521,7 +608,11 @@ private:
     Union_find<Facet> facets;
     { // fill the union find
       std::vector<Facet> non_filtered_facets;
-      tr_.incident_facets(v, std::back_inserter(non_filtered_facets));
+      if(tr_.is_parallel()) {
+	tr_.incident_facets_threadsafe(v, std::back_inserter(non_filtered_facets));
+      } else {
+	tr_.incident_facets(v, std::back_inserter(non_filtered_facets));
+      }
 
       for(typename std::vector<Facet>::iterator
             fit = non_filtered_facets.begin(),
@@ -747,6 +838,8 @@ public:
   operator>> (std::istream& is,
               Mesh_complex_3_in_triangulation_3_base<Tr2,Ct2> &c3t3);
 
+  void rescan_after_load_of_triangulation();
+
   static
   std::string io_signature()
   {
@@ -776,7 +869,11 @@ private:
   Triangulation tr_;
 
   typedef typename Base::Pair_of_vertices Pair_of_vertices;
+#ifdef CGAL_LINKED_WITH_TBB
+  typedef tbb::concurrent_hash_map<Pair_of_vertices, int> Edge_facet_counter;
+#else // not CGAL_LINKED_WITH_TBB
   typedef std::map<Pair_of_vertices, int> Edge_facet_counter;
+#endif // not CGAL_LINKED_WITH_TBB
 
   mutable Edge_facet_counter edge_facet_counter_;
 
@@ -809,7 +906,16 @@ Mesh_complex_3_in_triangulation_3_base<Tr,Ct>::add_to_complex(
         int edge_index_vb = tr_.vertex_triple_index(i, (j == 2) ? 0 : (j+1));
         Vertex_handle edge_va = cell->vertex(edge_index_va);
         Vertex_handle edge_vb = cell->vertex(edge_index_vb);
+#ifdef CGAL_LINKED_WITH_TBB
+	{
+	  typename Edge_facet_counter::accessor accessor;
+	  edge_facet_counter_.insert(accessor,
+				     this->make_ordered_pair(edge_va, edge_vb));
+	  ++accessor->second;
+	}
+#else // not CGAL_LINKED_WITH_TBB
         ++edge_facet_counter_[this->make_ordered_pair(edge_va, edge_vb)];
+#endif // not CGAL_LINKED_WITH_TBB
 
         const std::size_t n = edge_va->cached_number_of_incident_facets();
         const std::size_t m = edge_va->cached_number_of_components();
@@ -850,7 +956,16 @@ Mesh_complex_3_in_triangulation_3_base<Tr,Ct>::remove_from_complex(const Facet&
         const int edge_index_vb = tr_.vertex_triple_index(i, (j == 2) ? 0 : (j+1));
         const Vertex_handle edge_va = cell->vertex(edge_index_va);
         const Vertex_handle edge_vb = cell->vertex(edge_index_vb);
+#ifdef CGAL_LINKED_WITH_TBB
+	{
+	  typename Edge_facet_counter::accessor accessor;
+	  edge_facet_counter_.insert(accessor,
+				     this->make_ordered_pair(edge_va, edge_vb));
+	  --accessor->second;
+	}
+#else // not CGAL_LINKED_WITH_TBB
         --edge_facet_counter_[this->make_ordered_pair(edge_va, edge_vb)];
+#endif // not CGAL_LINKED_WITH_TBB
 
         const std::size_t n = edge_va->cached_number_of_incident_facets();
         CGAL_assertion(n>0);
@@ -920,27 +1035,35 @@ operator>> (std::istream& is,
     return is;
   }
 
+  c3t3.rescan_after_load_of_triangulation();
+  return is;
+}
+
+template <typename Tr, typename Ct>
+void
+Mesh_complex_3_in_triangulation_3_base<Tr,Ct>::
+rescan_after_load_of_triangulation() {
+  this->number_of_facets_ = 0;
   for(typename Tr::Finite_facets_iterator
-        fit = c3t3.triangulation().finite_facets_begin(),
-        end = c3t3.triangulation().finite_facets_end();
+        fit = this->triangulation().finite_facets_begin(),
+        end = this->triangulation().finite_facets_end();
       fit != end; ++fit)
   {
-    if ( c3t3.is_in_complex(*fit) ) {
-      ++c3t3.number_of_facets_;
+    if ( this->is_in_complex(*fit) ) {
+      ++this->number_of_facets_;
     }
   }
 
+  this->number_of_cells_ = 0;
   for(typename Tr::Finite_cells_iterator
-        cit = c3t3.triangulation().finite_cells_begin(),
-        end = c3t3.triangulation().finite_cells_end();
+        cit = this->triangulation().finite_cells_begin(),
+        end = this->triangulation().finite_cells_end();
       cit != end; ++cit)
   {
-    if ( c3t3.is_in_complex(cit) ) {
-      ++c3t3.number_of_cells_;
+    if ( this->is_in_complex(cit) ) {
+      ++this->number_of_cells_;
     }
   }
-
-  return is;
 }
 
 }  // end namespace Mesh_3
