@@ -27,6 +27,9 @@
 #ifndef CGAL_MESH_3_MESHER_3_H
 #define CGAL_MESH_3_MESHER_3_H
 
+#include <CGAL/license/Mesh_3.h>
+
+
 #include <CGAL/Mesh_3/config.h>
 
 #include <CGAL/Mesh_3/Dump_c3t3.h>
@@ -242,9 +245,6 @@ public:
   Mesher_status status() const;
 #endif
 
-private:
-  void remove_cells_from_c3t3();
-
 private:
   /// The oracle
   const MeshDomain& r_oracle_;
@@ -262,6 +262,16 @@ private:
   /// The container of the resulting mesh
   C3T3& r_c3t3_;
 
+  enum Refinement_stage {
+    NOT_INITIALIZED,
+    REFINE_FACETS,
+    REFINE_FACETS_AND_EDGES,
+    REFINE_FACETS_AND_EDGES_AND_VERTICES,
+    REFINE_ALL
+  };
+
+  Refinement_stage refinement_stage;
+
 private:
   // Disabled copy constructor
   Mesher_3(const Self& src);
@@ -317,7 +327,10 @@ Mesher_3<C3T3,MC,MD>::refine_mesh(std::string dump_after_refine_surface_prefix)
 
   // First surface mesh could modify c3t3 without notifying cells_mesher
   // So we have to ensure that no old cell will be left in c3t3
-  remove_cells_from_c3t3();
+  // Second, the c3t3 object could have been corrupted since the last call
+  // to `refine_mesh`, for example by inserting new vertices in the
+  // triangulation.
+  r_c3t3_.clear_cells_and_facets_from_c3t3();
 
 #ifndef CGAL_MESH_3_VERBOSE
   // Scan surface and refine it
@@ -328,6 +341,12 @@ Mesher_3<C3T3,MC,MD>::refine_mesh(std::string dump_after_refine_surface_prefix)
   WallClockTimer t;
 #endif
   facets_mesher_.refine(facets_visitor_);
+  facets_mesher_.scan_edges();
+  refinement_stage = REFINE_FACETS_AND_EDGES;
+  facets_mesher_.refine(facets_visitor_);
+  facets_mesher_.scan_vertices();
+  refinement_stage = REFINE_FACETS_AND_EDGES_AND_VERTICES;
+  facets_mesher_.refine(facets_visitor_);
 #ifdef CGAL_MESH_3_PROFILING
   double facet_ref_time = t.elapsed();
   std::cerr << "==== Facet refinement: " << facet_ref_time << " seconds ===="
@@ -359,6 +378,7 @@ Mesher_3<C3T3,MC,MD>::refine_mesh(std::string dump_after_refine_surface_prefix)
 
   // Then scan volume and refine it
   cells_mesher_.scan_triangulation();
+  refinement_stage = REFINE_ALL;
 #ifdef CGAL_MESH_3_PROFILING
   std::cerr << "Refining cells..." << std::endl;
   t.reset();
@@ -426,6 +446,7 @@ Mesher_3<C3T3,MC,MD>::refine_mesh(std::string dump_after_refine_surface_prefix)
   dump_c3t3(r_c3t3_, dump_after_refine_surface_prefix);
   std::cerr << "Start volume scan...";
   cells_mesher_.scan_triangulation();
+  refinement_stage = REFINE_ALL;
   std::cerr << "end scan. [Bad tets:" << cells_mesher_.size() << "]";
   std::cerr << std::endl << std::endl;
   elapsed_time += timer.time();
@@ -530,10 +551,10 @@ initialize()
 #  endif
       Random_points_on_sphere_3<Point> random_point(radius);
       const int NUM_PSEUDO_INFINITE_VERTICES = static_cast<int>(
-        tbb::task_scheduler_init::default_num_threads()
+        float(tbb::task_scheduler_init::default_num_threads())
         * Concurrent_mesher_config::get().num_pseudo_infinite_vertices_per_core);
       for (int i = 0 ; i < NUM_PSEUDO_INFINITE_VERTICES ; ++i, ++random_point)
-        r_c3t3_.add_far_point(*random_point + center);
+        r_c3t3_.add_far_point(r_c3t3_.triangulation().geom_traits().construct_translated_point_3_object()(*random_point, center));
     
 #  ifdef CGAL_CONCURRENT_MESH_3_VERBOSE
       std::cerr << "done." << std::endl;
@@ -548,6 +569,7 @@ initialize()
 
     // Scan triangulation
     facets_mesher_.scan_triangulation();
+    refinement_stage = REFINE_FACETS;
 
     // From now on, we're multi-thread
     r_c3t3_.triangulation().set_lock_data_structure(get_lock_data_structure());
@@ -607,7 +629,7 @@ initialize()
 
     // Scan triangulation
     facets_mesher_.scan_triangulation();
-
+    refinement_stage = REFINE_FACETS;
   }
 }
 
@@ -645,12 +667,25 @@ one_step()
 
     if ( facets_mesher_.is_algorithm_done() )
     {
-      facets_visitor_.activate();
-      cells_mesher_.scan_triangulation();
+      switch(refinement_stage) {
+      case REFINE_FACETS:
+        facets_mesher_.scan_edges();
+        refinement_stage = REFINE_FACETS_AND_EDGES;
+        break;
+      case REFINE_FACETS_AND_EDGES:
+        facets_mesher_.scan_vertices();
+        refinement_stage = REFINE_FACETS_AND_EDGES_AND_VERTICES;
+        break;
+      default:
+        facets_visitor_.activate();
+        cells_mesher_.scan_triangulation();
+        refinement_stage = REFINE_ALL;
+      }
     }
   }
   else
   {
+    CGAL_assertion(refinement_stage == REFINE_ALL);
     cells_mesher_.one_step(cells_visitor_);
   }
 }
@@ -676,20 +711,6 @@ status() const
 }
 #endif
 
-
-template<class C3T3, class MC, class MD>
-void
-Mesher_3<C3T3,MC,MD>::
-remove_cells_from_c3t3()
-{
-  for ( typename C3T3::Triangulation::Finite_cells_iterator
-    cit = r_c3t3_.triangulation().finite_cells_begin(),
-    end = r_c3t3_.triangulation().finite_cells_end() ; cit != end ; ++cit )
-  {
-    r_c3t3_.remove_from_complex(cit);
-  }
-}
-
 template<class C3T3, class MC, class MD>
 inline
 std::string
