@@ -19,6 +19,9 @@
 #ifndef CGAL_SURFACE_MESH_H
 #define CGAL_SURFACE_MESH_H
 
+#include <CGAL/license/Surface_mesh.h>
+
+
 #include <iterator>
 #include <algorithm>
 #include <utility>
@@ -33,17 +36,19 @@
 #include <boost/array.hpp>
 #include <boost/iterator/iterator_facade.hpp>
 #include <boost/foreach.hpp>
-#include <boost/property_map/property_map.hpp>
 
+#include <CGAL/property_map.h>
 #include <CGAL/Iterator_range.h>
 #include <CGAL/circulator.h>
 #include <CGAL/assertions.h>
 #include <CGAL/Surface_mesh/Surface_mesh_fwd.h>
 #include <CGAL/Surface_mesh/IO.h>
-//#include <CGAL/Surface_mesh/Properties.h>
+#include <CGAL/Surface_mesh/Properties.h>
 #include <CGAL/boost/graph/graph_traits_Surface_mesh.h>
 #include <CGAL/boost/graph/iterator.h>
 #include <CGAL/boost/graph/Euler_operations.h>
+#include <CGAL/IO/File_scanner_OFF.h>
+#include <CGAL/Handle_hash_function.h>
 
 namespace CGAL {
 
@@ -195,8 +200,7 @@ namespace CGAL {
 
         SM_Edge_index(size_type idx) : halfedge_(idx * 2) { }
 
-
-        SM_Edge_index(SM_Halfedge_index he) : halfedge_(he) { }
+        explicit SM_Edge_index(SM_Halfedge_index he) : halfedge_(he) { }
 
         // returns the internal halfedge.
         SM_Halfedge_index halfedge() const { return halfedge_; }
@@ -280,492 +284,22 @@ class Surface_mesh
     class Handle_iterator;
 public:
 
-
-
-/// \addtogroup PkgSurface_mesh
-///
-/// @{
-
-/// @cond CGAL_DOCUMENT_INTERNALS
-class Base_property_array
-{
-public:
-
-    /// Default constructor
-    Base_property_array(const std::string& name) : name_(name) {}
-
-    /// Destructor.
-    virtual ~Base_property_array() {}
-
-    /// Reserve memory for n elements.
-    virtual void reserve(size_t n) = 0;
-
-    /// Resize storage to hold n elements.
-    virtual void resize(size_t n) = 0;
-
-    /// Free unused memory.
-    virtual void shrink_to_fit() = 0;
-
-    /// Extend the number of elements by one.
-    virtual void push_back() = 0;
-
-    virtual bool transfer(const Base_property_array& other) = 0;
-
-    /// Let two elements swap their storage place.
-    virtual void swap(size_t i0, size_t i1) = 0;
-
-    /// Return a deep copy of self.
-    virtual Base_property_array* clone () const = 0;
-
-    /// Return the type_info of the property
-    virtual const std::type_info& type() = 0;
-
-    /// Return the name of the property
-    const std::string& name() const { return name_; }
-
-
-protected:
-
-    std::string name_;
-};
-
-  /// @endcond
-
-
-//== CLASS DEFINITION =========================================================
-
-/// @cond CGAL_DOCUMENT_INTERNALS
-
-template <class T>
-class Property_array : public Base_property_array
-{
-public:
-
-    typedef T                                       value_type;
-    typedef std::vector<value_type>                 vector_type;
-    typedef typename vector_type::reference         reference;
-    typedef typename vector_type::const_reference   const_reference;
-
-    Property_array(const std::string& name, T t=T()) : Base_property_array(name), value_(t) {}
-
-public: // virtual interface of Base_property_array
-
-    virtual void reserve(size_t n)
-    {
-        data_.reserve(n);
-    }
-
-    virtual void resize(size_t n)
-    {
-        data_.resize(n, value_);
-    }
-
-    virtual void push_back()
-    {
-        data_.push_back(value_);
-    }
-
-    bool transfer(const Base_property_array& other)
-    {
-      const Property_array<T>* pa = dynamic_cast<const Property_array*>(&other);
-      if(pa != NULL){
-        std::copy((*pa).data_.begin(), (*pa).data_.end(), data_.end()-(*pa).data_.size());
-        return true;
-      } 
-      return false;
-    }
-
-    virtual void shrink_to_fit()
-    {
-        vector_type(data_).swap(data_);
-    }
-
-    virtual void swap(size_t i0, size_t i1)
-    {
-        T d(data_[i0]);
-        data_[i0]=data_[i1];
-        data_[i1]=d;
-    }
-
-    virtual Base_property_array* clone() const
-    {
-        Property_array<T>* p = new Property_array<T>(this->name_, this->value_);
-        p->data_ = data_;
-        return p;
-    }
-
-    virtual const std::type_info& type() { return typeid(T); }
-
-
-public:
-
-    /// Get pointer to array (does not work for T==bool)
-    const T* data() const
-    {
-        return &data_[0];
-    }
-
-    /// Access the i'th element. No range check is performed!
-    reference operator[](int _idx)
-    {
-        CGAL_assertion( size_t(_idx) < data_.size() );
-        return data_[_idx];
-    }
-
-    /// Const access to the i'th element. No range check is performed!
-    const_reference operator[](int _idx) const
-    {
-        CGAL_assertion( size_t(_idx) < data_.size());
-        return data_[_idx];
-    }
-
-
-
-private:
-    vector_type data_;
-    value_type  value_;
-};
-
-
-#if 0
-// specialization for bool properties
-template <>
-inline const bool*
-Property_array<bool>::data() const
-{
-    CGAL_assertion(false);
-    return NULL;
-}
-#endif 
-
-  /// @endcond
-
-//== CLASS DEFINITION =========================================================
-
-/// @cond CGAL_DOCUMENT_INTERNALS
-
-template<typename>
-class Property_container;
-/// @endcond 
-
-
-
-
-//== CLASS DEFINITION =========================================================
-/// @cond CGAL_DOCUMENT_INTERNALS
-
-template <class, class>
-class Property_map;
-
-template<typename Key>
-class Property_container
-{
-public:
-
-    // default constructor
-    Property_container() : size_(0) {}
-
-    // destructor (deletes all property arrays)
-    virtual ~Property_container() { clear(); }
-
-    // copy constructor: performs deep copy of property arrays
-    Property_container(const Property_container& _rhs) { operator=(_rhs); }
-
-    // assignment: performs deep copy of property arrays
-    Property_container& operator=(const Property_container& _rhs)
-    {
-        if (this != &_rhs)
-        {
-            clear();
-            parrays_.resize(_rhs.n_properties());
-            size_ = _rhs.size();
-            for (unsigned int i=0; i<parrays_.size(); ++i)
-                parrays_[i] = _rhs.parrays_[i]->clone();
-        }
-        return *this;
-    }
-
-    void transfer(const Property_container& _rhs)
-    {
-      for(unsigned int i=0; i<parrays_.size(); ++i){
-        for (unsigned int j=0; j<_rhs.parrays_.size(); ++j){
-          if(parrays_[i]->name() ==  _rhs.parrays_[j]->name()){
-            parrays_[i]->transfer(* _rhs.parrays_[j]);
-            break;
-          }
-        }
-      }
-    }
-
-    // returns the current size of the property arrays
-    size_t size() const { return size_; }
-
-    // returns the number of property arrays
-    size_t n_properties() const { return parrays_.size(); }
-
-    // returns a vector of all property names
-    std::vector<std::string> properties() const
-    {
-        std::vector<std::string> names;
-        for (unsigned int i=0; i<parrays_.size(); ++i)
-            names.push_back(parrays_[i]->name());
-        return names;
-    }
-
-    // add a property with name \c name and default value \c t
-    template <class T>
-    std::pair<Property_map<Key, T>, bool>
-    add(const std::string& name, const T t=T())
-    {
-        for (unsigned int i=0; i<parrays_.size(); ++i)
-        {
-            if (parrays_[i]->name() == name)
-            {
-              return std::make_pair(Property_map<Key, T>(dynamic_cast<Property_array<T>*>(parrays_[i])), false);
-            }
-        }
-
-        // otherwise add the property
-        Property_array<T>* p = new Property_array<T>(name, t);
-        p->resize(size_);
-        parrays_.push_back(p);
-        return std::make_pair(Property_map<Key, T>(p), true);
-    }
-
-
-    // get a property by its name. returns invalid property if it does not exist.
-    template <class T> 
-    std::pair<Property_map<Key, T>,bool>
-    get(const std::string& name) const
-    {
-        for (unsigned int i=0; i<parrays_.size(); ++i)
-            if (parrays_[i]->name() == name)
-              return std::make_pair(Property_map<Key, T>(dynamic_cast<Property_array<T>*>(parrays_[i])), true);
-        return std::make_pair(Property_map<Key, T>(), false);
-    }
-
-
-    // returns a property if it exists, otherwise it creates it first.
-    template <class T>
-    Property_map<Key, T> 
-    get_or_add(const std::string& name, const T t=T())
-    {
-      Property_map<Key, T> p;
-      bool b;
-      boost::tie(p,b)= get<T>(name);
-        if (!b) p = add<T>(name, t).first;
-        return p;
-    }
-
-
-    // get the type of property by its name. returns typeid(void) if it does not exist.
-    const std::type_info& 
-    get_type(const std::string& name)
-    {
-        for (unsigned int i=0; i<parrays_.size(); ++i)
-            if (parrays_[i]->name() == name)
-                return parrays_[i]->type();
-        return typeid(void);
-    }
-
-
-    // delete a property
-    template <class T> 
-    void
-    remove(Property_map<Key, T>& h)
-    {
-        typename std::vector<Base_property_array*>::iterator it=parrays_.begin(), end=parrays_.end();
-        for (; it!=end; ++it)
-        {
-            if (*it == h.parray_)
-            {
-                delete *it;
-                parrays_.erase(it);
-                h.reset();
-                break;
-            }
-        }
-    }
-
-
-    // delete all properties
-    void clear()
-    {
-        for (unsigned int i=0; i<parrays_.size(); ++i)
-            delete parrays_[i];
-        parrays_.clear();
-        size_ = 0;
-    }
-
-
-    // reserve memory for n entries in all arrays
-    void reserve(size_t n) const
-    {
-        for (unsigned int i=0; i<parrays_.size(); ++i)
-            parrays_[i]->reserve(n);
-    }
-
-    // resize all arrays to size n
-    void resize(size_t n)
-    {
-        for (unsigned int i=0; i<parrays_.size(); ++i)
-            parrays_[i]->resize(n);
-        size_ = n;
-    }
-
-    // free unused space in all arrays
-    void shrink_to_fit() const
-    {
-        for (unsigned int i=0; i<parrays_.size(); ++i)
-            parrays_[i]->shrink_to_fit();
-    }
-
-    // add a new element to each vector
-    void push_back()
-    {
-        for (unsigned int i=0; i<parrays_.size(); ++i)
-            parrays_[i]->push_back();
-        ++size_;
-    }
-
-    // swap elements i0 and i1 in all arrays
-    void swap(size_t i0, size_t i1) const
-    {
-        for (unsigned int i=0; i<parrays_.size(); ++i)
-            parrays_[i]->swap(i0, i1);
-    }
-
-
-private:
-    std::vector<Base_property_array*>  parrays_;
-    size_t  size_;
-};
-
-  /// @endcond
-
 #ifndef DOXYGEN_RUNNING
-/// 
-///
-/// `Property_map` enables to attach properties to the simplices of a 
-///  surface mesh.
-/// 
-/// @tparam Key The key type of the property map. It must be a model of `Index`.
-/// @tparam Value The value type of the property.
-///
-/// \cgalModels `LvaluePropertyMap`
-///
-template <class I, class T>
-class Property_map
-/// @cond CGAL_DOCUMENT_INTERNALS
-  : public boost::put_get_helper< 
-           typename Property_array<T>::reference,
-           Property_map< I, T > >
-/// @endcond
-{
-    typedef void (Property_map::*bool_type)() const;
-    void this_type_does_not_support_comparisons() const {}
-public:
-    typedef I key_type;
-    typedef T value_type;
-    typedef boost::lvalue_property_map_tag category;
-
-#ifndef DOXYGEN_RUNNING
-
-    typedef typename Property_array<T>::reference reference;
-
-    typedef typename Property_array<T>::const_reference const_reference;
-#else 
-    /// A reference to the value type of the property.
-  typedef unspecified_type reference;
-
-    /// A const reference to the value type of the property.
-  typedef unspecified_type const_reference;
-#endif
-
-#ifndef DOXYGEN_RUNNING
-    friend class Property_container<I>;
-
-    template <typename K>  friend class Surface_mesh;
-#endif
-
-public:
-/// @cond CGAL_DOCUMENT_INTERNALS
-    Property_map(Property_array<T>* p=NULL) : parray_(p) {}
-
-    void reset()
-    {
-        parray_ = NULL;
-    }
-  /// @endcond 
-
-public:
-    /// \name Accessing Properties
-    //@{
-#ifdef DOXYGEN_RUNNING
-    /// Conversion to a Boolean. It is \c true when the property map
-    /// can be used, and \c false otherwise.  
-  operator bool () const;
-#else
-    operator bool_type() const {
-        return parray_ != NULL ?
-            &Property_map::this_type_does_not_support_comparisons : 0;
-    }
-#endif
-    /// Access the property associated with the key \c i.
-    reference operator[](const I& i)
-    {
-      CGAL_assertion(parray_ != NULL);
-      return (*parray_)[i];
-    }
-
-    /// Access the property associated with the key \c i.
-    reference operator[](const I& i) const
-    {
-      CGAL_assertion(parray_ != NULL);
-      return (*parray_)[i];
-    }
-
-    bool transfer (const Property_map& other)
-    {
-      return parray_->transfer(*(other.parray_));
-    }
-
-    /// Allows access to the underlying storage of the property. This
-    /// is useful when the key associated with the properties is
-    /// unimportant and only the properties are of interest
-    /// (e.g. rendering).
-    ///
-    /// \returns a pointer to the underlying storage of the property.
-    const T* data() const
-    {
-      CGAL_assertion(parray_ != NULL);
-      return parray_->data();
-    }
-
-    //@}
-private:
-
-    Property_array<T>& array()
-    {
-        CGAL_assertion(parray_ != NULL);
-        return *parray_;
-    }
-
-    const Property_array<T>& array() const
-    {
-        CGAL_assertion(parray_ != NULL);
-        return *parray_;
-    }
-
-    Property_array<T>* parray_;
-};
+  template <class I, class T>
+  struct Property_map : Properties::Property_map_base<I, T, Property_map<I, T> >
+  {
+    typedef Properties::Property_map_base<I, T, Property_map<I, T> > Base;
+    typedef typename Base::reference reference;
+    Property_map() : Base() {}
+    Property_map(const Base& pm): Base(pm) {}
+  };
 
+  template <typename Key, typename T>
+  struct Get_property_map {
+    typedef Property_map<Key, T> type;
+  };
 #endif // DOXYGEN_RUNNING
 
-///@}
-
-
-
     /// \name Basic Types
     ///
     ///@{
@@ -2149,7 +1683,7 @@ public:
     ///
     ///  A halfedge, or edge is on the border of a surface mesh
     /// if it is incident to a `null_face()`.  A vertex is on a border
-    /// if it is incident to a border halfedge. While for a halfedge and
+    /// if it is isolated or incident to a border halfedge. While for a halfedge and
     /// edge this is a constant time operation, for a vertex it means
     /// to look at all incident halfedges.  If algorithms operating on a 
     /// surface mesh maintain that the halfedge associated to a border vertex is
@@ -2163,13 +1697,13 @@ public:
     /// With the default value for
     /// `check_all_incident_halfedges` the function iteratates over the incident halfedges.
     /// With `check_all_incident_halfedges == false` the function returns `true`, if the incident
-    /// halfedge associated to vertex `v` is a border halfedge.
+    /// halfedge associated to vertex `v` is a border halfedge, or if the vertex is isolated.
     /// \cgalAdvancedEnd
   bool is_border(Vertex_index v, bool check_all_incident_halfedges = true) const
     {
         Halfedge_index h(halfedge(v));
         if (h == null_halfedge()){
-          return false;
+          return true;
         }
         if(check_all_incident_halfedges){
           Halfedge_around_target_circulator hatc(h,*this), done(hatc);
@@ -2180,7 +1714,7 @@ public:
           }while(++hatc != done);
           return false;
         }
-        return (!(is_valid(h) && is_border(h)));
+        return is_border(h);
     }
 
     /// returns whether `h` is a border halfege, that is if its incident face is `sm.null_face()`.
@@ -2264,28 +1798,32 @@ private: //--------------------------------------------------- property handling
   struct Property_selector<typename CGAL::Surface_mesh<P>::Vertex_index, dummy> {
     CGAL::Surface_mesh<P>* m_;
     Property_selector(CGAL::Surface_mesh<P>* m) : m_(m) {}
-    Property_container<typename CGAL::Surface_mesh<P>::Vertex_index>&
+    Properties::Property_container<Self,
+                                   typename CGAL::Surface_mesh<P>::Vertex_index>&
     operator()() { return m_->vprops_; }
   };
   template<bool dummy>
   struct Property_selector<typename CGAL::Surface_mesh<P>::Halfedge_index, dummy> {
     CGAL::Surface_mesh<P>* m_;
     Property_selector(CGAL::Surface_mesh<P>* m) : m_(m) {}
-    Property_container<typename CGAL::Surface_mesh<P>::Halfedge_index>&
+    Properties::Property_container<Self,
+                                   typename CGAL::Surface_mesh<P>::Halfedge_index>&
     operator()() { return m_->hprops_; }
   };
   template<bool dummy>
   struct Property_selector<typename CGAL::Surface_mesh<P>::Edge_index, dummy> {
     CGAL::Surface_mesh<P>* m_;
     Property_selector(CGAL::Surface_mesh<P>* m) : m_(m) {}
-    Property_container<typename CGAL::Surface_mesh<P>::Edge_index>&
+    Properties::Property_container<Self,
+                                   typename CGAL::Surface_mesh<P>::Edge_index>&
     operator()() { return m_->eprops_; }
   };
   template<bool dummy>
   struct Property_selector<typename CGAL::Surface_mesh<P>::Face_index, dummy> {
     CGAL::Surface_mesh<P>* m_;
     Property_selector(CGAL::Surface_mesh<P>* m) : m_(m) {}
-    Property_container<typename CGAL::Surface_mesh<P>::Face_index>&
+    Properties::Property_container<Self,
+                                   typename CGAL::Surface_mesh<P>::Face_index>&
     operator()() { return m_->fprops_; }
   };
 
@@ -2294,7 +1832,7 @@ private: //--------------------------------------------------- property handling
 
  /*! \name Property Handling
 
- A `Property_map<I,T>` allows to associate properties of type `T` to a vertex, halfdge, edge, or face index type I.
+ A `Properties::Property_map<I,T>` allows to associate properties of type `T` to a vertex, halfdge, edge, or face index type I.
  Properties can be added, and looked up with a string, and they can be removed at runtime.
  The \em point property of type `P` is associated to the string "v:point". 
 
@@ -2341,7 +1879,7 @@ private: //--------------------------------------------------- property handling
     template<class I, class T>
     void remove_property_map(Property_map<I, T>& p)
     {
-      (Property_selector<I>(this)()).remove(p);
+      (Property_selector<I>(this)()).template remove<T>(p);
     }
 
     /// @cond CGAL_DOCUMENT_INTERNALS
@@ -2363,13 +1901,16 @@ private: //--------------------------------------------------- property handling
     template<class I>
     std::vector<std::string> properties() const
     {
-      return Property_selector<I>(this)().properties();
+      return Property_selector<I>(const_cast<Self*>(this))().properties();
     }
 
     /// returns the property for the string "v:point".
     Property_map<Vertex_index, Point>
     points() const { return vpoint_; }
 
+    Property_map<Vertex_index, Point>&
+    points() { return vpoint_; }
+
     /// returns the point associated to vertex `v`.
     const Point&
     point(Vertex_index v) const { return vpoint_[v]; }
@@ -2422,10 +1963,10 @@ private: //--------------------------------------------------- helper functions
     void adjust_incoming_halfedge(Vertex_index v);
 
 private: //------------------------------------------------------- private data
-    Property_container<Vertex_index> vprops_;
-    Property_container<Halfedge_index> hprops_;
-    Property_container<Edge_index> eprops_;
-    Property_container<Face_index> fprops_;
+    Properties::Property_container<Self, Vertex_index> vprops_;
+    Properties::Property_container<Self, Halfedge_index> hprops_;
+    Properties::Property_container<Self, Edge_index> eprops_;
+    Properties::Property_container<Self, Face_index> fprops_;
 
     Property_map<Vertex_index, Vertex_connectivity>      vconn_;
     Property_map<Halfedge_index, Halfedge_connectivity>  hconn_;
@@ -2502,41 +2043,90 @@ private: //------------------------------------------------------- private data
   inline std::istream& sm_skip_comments( std::istream& in) {
       char c;
       in >> c;
-      if (c == '#')
-        in.ignore((std::numeric_limits<std::streamsize>::max)(), '\n');
-      else
-        in.putback(c);
+      while(c == '#')
+      {
+       in.ignore((std::numeric_limits<std::streamsize>::max)(), '\n');
+       in >> c;
+      }
+      in.putback(c);
       return in;
   }
 /// @endcond
 
 
   /// \relates Surface_mesh
-  /// Extracts the surface mesh from an input stream in Ascii OFF format.
-  /// The operator only reads the point property and does not read files 
-  /// with vertex normals or textures.
+
+  /// \relates Surface_mesh
+  /// Extracts the surface mesh from an input stream in Ascii OFF, COFF, NOFF, CNOFF format.
+  /// The operator reads the point property as well as "v:normal", "v:color", and "f:color".
+  /// Vertex texture coordinates are ignored.
   /// \pre `operator>>(std::istream&,const P&)` must be defined.
   template <typename P>
   std::istream& operator>>(std::istream& is, Surface_mesh<P>& sm)
   {
-    typedef Surface_mesh<P> Mesh;
-    typedef typename Mesh::size_type size_type;
+   typedef Surface_mesh<P> Mesh;
+   typedef typename Kernel_traits<P>::Kernel K;
+   typedef typename K::Vector_3 Vector_3;
+   typedef typename Mesh::Face_index Face_index;
+   typedef typename Mesh::Vertex_index Vertex_index;
+   typedef typename Mesh::size_type size_type;
     sm.clear();
     int n, f, e;
     std::string off;
+    is >> sm_skip_comments;
     is >> off;
-    CGAL_assertion(off == "OFF" || off == "COFF");
+    CGAL_assertion( (off == "OFF") || (off == "COFF") || (off == "NOFF") || (off == "CNOFF"));
+
     is >> n >> f >> e;
+
     sm.reserve(n,2*f,e);
     P p;
+    Vector_3 v;
+    typename Mesh::template Property_map<Vertex_index,CGAL::Color> vcolor;
+    typename Mesh::template Property_map<Vertex_index,Vector_3> vnormal;
+    bool vcolored = false, v_has_normals = false;
+
+    if((off == "NOFF") || (off == "CNOFF")){
+      bool created;
+      boost::tie(vnormal, created) = sm.template add_property_map<Vertex_index,Vector_3>("v:normal",Vector_3(0,0,0));
+      v_has_normals = true;
+    }
+    char ci;
+
     for(int i=0; i < n; i++){
       is >> sm_skip_comments;
       is >> p;
-      sm.add_vertex(p);
-      is.ignore((std::numeric_limits<std::streamsize>::max)(), '\n');
+      Vertex_index vi= sm.add_vertex(p);
+      if(v_has_normals){
+        is >> v;
+        vnormal[vi] = v;
+      }
+
+
+      if(i == 0 && ((off == "COFF") || (off == "CNOFF"))){
+        std::string col;
+        std::getline(is, col);
+        std::istringstream iss(col);
+        if(iss >> ci){
+         bool created;
+         boost::tie(vcolor, created) = sm.template add_property_map<Vertex_index,CGAL::Color>("v:color",CGAL::Color(0,0,0));
+         std::istringstream iss2(col);
+         vcolor[vi] = File_scanner_OFF::get_color_from_line(iss2);
+         vcolored = true;
+        }
+      }else{
+         if(vcolored){
+           //stores the RGB value
+           vcolor[vi] = File_scanner_OFF::get_color_from_line(is);
+         }
+       }
     }
     std::vector<size_type> vr;
     std::size_t d;
+
+    bool fcolored = false;
+    typename Mesh::template Property_map<Face_index,CGAL::Color> fcolor;
+
     for(int i=0; i < f; i++){
       is >> sm_skip_comments;
       is >> d;
@@ -2544,7 +2134,32 @@ private: //------------------------------------------------------- private data
       for(std::size_t j=0; j<d; j++){
         is >> vr[j];
       }
-      sm.add_face(vr);
+      Face_index fi = sm.add_face(vr);
+      if(fi == sm.null_face())
+      {
+       std::cout<< "Warning: Facets don't seem to be oriented. Loading a Soup of polygons instead."<<std::endl;
+       sm.clear();
+       return is;
+      }
+
+      // the first face will tell us if faces have a color map
+      // TODO: extend this to RGBA
+      if(i == 0 ){
+        std::string col;
+        std::getline(is, col);
+        std::istringstream iss(col);
+        if(iss >> ci){
+          bool created;
+          boost::tie(fcolor, created) = sm.template add_property_map<Face_index,CGAL::Color>("f:color",CGAL::Color(0,0,0));
+          fcolored = true;
+          std::istringstream iss2(col);
+          fcolor[fi] = File_scanner_OFF::get_color_from_line(iss2);
+        }
+      } else {
+          if(fcolored){
+            fcolor[fi] = File_scanner_OFF::get_color_from_line(is);
+          }
+        }
     }
     return is;
   }
@@ -2961,7 +2576,45 @@ collect_garbage()
     garbage_ = false;
 }
 
+namespace internal{
+  namespace handle {
+    template <>
+    struct Hash_functor<SM_Vertex_index>{
+      std::size_t
+      operator()(const SM_Vertex_index i)
+      {
+        return i;
+      }
+    };
 
+    template <>
+    struct Hash_functor<SM_Halfedge_index>{
+      std::size_t
+      operator()(const SM_Halfedge_index i)
+      {
+        return i;
+      }
+    };
+
+    template <>
+    struct Hash_functor<SM_Edge_index>{
+      std::size_t
+      operator()(const SM_Edge_index i)
+      {
+        return i;
+      }
+    };
+
+    template <>
+    struct Hash_functor<SM_Face_index>{
+      std::size_t
+      operator()(const SM_Face_index i)
+      {
+        return i;
+      }
+    };
+  }
+}
 
 } // CGAL
 
